@@ -15,6 +15,7 @@
 
 """Storage backend for S3 or Storage Servers that follow the S3 Protocol"""
 
+from concurrent import futures
 import io
 import logging
 import math
@@ -32,7 +33,6 @@ except ImportError:
     boto_exceptions = None
     boto_utils = None
 
-import eventlet
 from oslo_config import cfg
 from oslo_utils import encodeutils
 from oslo_utils import units
@@ -465,7 +465,8 @@ class Store(glance_store.driver.Store):
 
     def _option_get(self, param):
         if self.backend_group:
-            store_conf = getattr(self.conf, self.backend_group)
+            store_conf = glance_store.driver.BackendGroupConfiguration(
+                self.OPTIONS, self.backend_group, conf=self.conf)
         else:
             store_conf = self.conf.glance_store
 
@@ -648,7 +649,8 @@ class Store(glance_store.driver.Store):
                                             key=key,
                                             loc=loc,
                                             hashing_algo=hashing_algo,
-                                            verifier=verifier)
+                                            verifier=verifier,
+                                            image_size=image_size)
 
             return self._add_multipart(s3_client=s3_client,
                                        image_file=image_file,
@@ -663,7 +665,7 @@ class Store(glance_store.driver.Store):
         raise exceptions.Duplicate(image=key)
 
     def _add_singlepart(self, s3_client, image_file, bucket, key, loc,
-                        hashing_algo, verifier):
+                        hashing_algo, verifier, image_size):
         """Stores an image file with a single part upload to S3 backend.
 
         :param s3_client: An object with credentials to connect to S3
@@ -674,6 +676,8 @@ class Store(glance_store.driver.Store):
                     from glance_store.location.get_location_from_uri()
         :param hashing_algo: A hashlib algorithm identifier (string)
         :param verifier: An object used to verify signatures for images
+        :param image_size: The size of the image data to write,
+                           in bytes
         :returns: tuple of: (1) URL in backing store, (2) bytes written,
                   (3) checksum, (4) multihash value, and (5) a dictionary
                   with storage system specific information
@@ -681,15 +685,28 @@ class Store(glance_store.driver.Store):
         os_hash_value = utils.get_hasher(hashing_algo, False)
         checksum = utils.get_hasher('md5', False)
         image_data = b''
-        image_size = 0
+        total_bytes = 0
         for chunk in utils.chunkreadable(image_file, self.WRITE_CHUNKSIZE):
+            total_bytes += len(chunk)
+            if image_size != 0 and total_bytes > image_size:
+                # Size exceeded, abort upload
+                raise glance_store.Invalid(
+                    _("Size exceeds: expected %(expected)d bytes, "
+                      "got %(actual)d bytes") %
+                    {'expected': image_size, 'actual': total_bytes})
             image_data += chunk
-            image_size += len(chunk)
             os_hash_value.update(chunk)
             checksum.update(chunk)
             if verifier:
                 verifier.update(chunk)
 
+        # Final size check after reading all chunks
+        if image_size != 0 and total_bytes != image_size:
+            raise glance_store.Invalid(
+                _("Size mismatch: expected %(expected)d bytes, "
+                  "got %(actual)d bytes") %
+                {'expected': image_size, 'actual': total_bytes})
+
         s3_client.put_object(Body=image_data,
                              Bucket=bucket,
                              Key=key)
@@ -703,9 +720,9 @@ class Store(glance_store.driver.Store):
 
         LOG.debug("Wrote %(size)d bytes to S3 key named %(key)s "
                   "with checksum %(checksum)s",
-                  {'size': image_size, 'key': key, 'checksum': checksum_hex})
+                  {'size': total_bytes, 'key': key, 'checksum': checksum_hex})
 
-        return loc.get_uri(), image_size, checksum_hex, hash_hex, metadata
+        return loc.get_uri(), total_bytes, checksum_hex, hash_hex, metadata
 
     def _add_multipart(self, s3_client, image_file, image_size, bucket,
                        key, loc, hashing_algo, verifier):
@@ -726,71 +743,98 @@ class Store(glance_store.driver.Store):
         os_hash_value = utils.get_hasher(hashing_algo, False)
         checksum = utils.get_hasher('md5', False)
         pool_size = self.s3_store_thread_pools
-        pool = eventlet.greenpool.GreenPool(size=pool_size)
-        mpu = s3_client.create_multipart_upload(Bucket=bucket, Key=key)
-        upload_id = mpu['UploadId']
-        LOG.debug("Multipart initiate key=%(key)s, UploadId=%(UploadId)s",
-                  {'key': key, 'UploadId': upload_id})
-        cstart = 0
-        plist = []
+        # Replace eventlet.GreenPool with ThreadPoolExecutor
+        with futures.ThreadPoolExecutor(
+                max_workers=pool_size) as executor:
+            # Create a list to store the futures
+            futures_list = []
+            mpu = s3_client.create_multipart_upload(Bucket=bucket, Key=key)
+            upload_id = mpu['UploadId']
+            LOG.debug("Multipart initiate key=%(key)s, UploadId=%(UploadId)s",
+                      {'key': key, 'UploadId': upload_id})
+            cstart = 0
+            plist = []
 
-        chunk_size = int(math.ceil(float(image_size) / MAX_PART_NUM))
-        write_chunk_size = max(self.s3_store_large_object_chunk_size,
-                               chunk_size)
-        it = utils.chunkreadable(image_file, self.WRITE_CHUNKSIZE)
-        buffered_chunk = b''
-        while True:
-            try:
-                buffered_clen = len(buffered_chunk)
-                if buffered_clen < write_chunk_size:
-                    # keep reading data
-                    read_chunk = next(it)
-                    buffered_chunk += read_chunk
-                    continue
-                else:
-                    write_chunk = buffered_chunk[:write_chunk_size]
-                    remained_data = buffered_chunk[write_chunk_size:]
-                    os_hash_value.update(write_chunk)
-                    checksum.update(write_chunk)
-                    if verifier:
-                        verifier.update(write_chunk)
-                    fp = io.BytesIO(write_chunk)
-                    fp.seek(0)
-                    part = UploadPart(mpu, fp, cstart + 1, len(write_chunk))
-                    pool.spawn_n(run_upload, s3_client, bucket, key, part)
-                    plist.append(part)
-                    cstart += 1
-                    buffered_chunk = remained_data
-            except StopIteration:
-                if len(buffered_chunk) > 0:
-                    # Write the last chunk data
-                    write_chunk = buffered_chunk
-                    os_hash_value.update(write_chunk)
-                    checksum.update(write_chunk)
-                    if verifier:
-                        verifier.update(write_chunk)
-                    fp = io.BytesIO(write_chunk)
-                    fp.seek(0)
-                    part = UploadPart(mpu, fp, cstart + 1, len(write_chunk))
-                    pool.spawn_n(run_upload, s3_client, bucket, key, part)
-                    plist.append(part)
-                break
+            chunk_size = int(math.ceil(float(image_size) / MAX_PART_NUM))
+            write_chunk_size = max(self.s3_store_large_object_chunk_size,
+                                   chunk_size)
+            it = utils.chunkreadable(image_file, self.WRITE_CHUNKSIZE)
+            buffered_chunk = b''
+            total_bytes = 0
+            while True:
+                try:
+                    buffered_clen = len(buffered_chunk)
+                    if buffered_clen < write_chunk_size:
+                        # keep reading data
+                        read_chunk = next(it)
+                        buffered_chunk += read_chunk
+                        continue
+                    else:
+                        write_chunk = buffered_chunk[:write_chunk_size]
+                        remained_data = buffered_chunk[write_chunk_size:]
+                        total_bytes += len(write_chunk)
+                        if image_size != 0 and total_bytes > image_size:
+                            # Abort multipart upload immediately
+                            s3_client.abort_multipart_upload(
+                                Bucket=bucket, Key=key, UploadId=upload_id)
+                            raise glance_store.Invalid(
+                                _("Size exceeds: expected "
+                                  "%(expected)d "
+                                  "bytes, got %(actual)d bytes") %
+                                {'expected': image_size,
+                                 'actual': total_bytes})
+                        os_hash_value.update(write_chunk)
+                        checksum.update(write_chunk)
+                        if verifier:
+                            verifier.update(write_chunk)
+                        fp = io.BytesIO(write_chunk)
+                        fp.seek(0)
+                        part = UploadPart(
+                            mpu, fp, cstart + 1, len(write_chunk))
+                        # Spawn thread to upload part
+                        futures_list.append(executor.submit(
+                            run_upload, s3_client, bucket, key, part))
+                        plist.append(part)
+                        cstart += 1
+                        buffered_chunk = remained_data
+                except StopIteration:
+                    if len(buffered_chunk) > 0:
+                        # Write the last chunk data
+                        write_chunk = buffered_chunk
+                        total_bytes += len(write_chunk)
+                        os_hash_value.update(write_chunk)
+                        checksum.update(write_chunk)
+                        if verifier:
+                            verifier.update(write_chunk)
+                        fp = io.BytesIO(write_chunk)
+                        fp.seek(0)
+                        part = UploadPart(
+                            mpu, fp, cstart + 1, len(write_chunk))
+                        futures_list.append(executor.submit(
+                            run_upload, s3_client, bucket, key, part))
+                        plist.append(part)
+                    break
 
-        pedict = {}
-        total_size = 0
-        pool.waitall()
+        # Wait for all uploads to finish
+        futures.wait(futures_list)
 
-        for part in plist:
-            pedict.update(part.etag)
-            total_size += part.size
+        # Final size validation after all parts uploaded
+        if image_size != 0 and total_bytes != image_size:
+            # Abort upload if mismatch (redundant check)
+            s3_client.abort_multipart_upload(
+                Bucket=bucket, Key=key, UploadId=upload_id)
+            raise glance_store.Invalid(
+                _("Size mismatch: expected %(expected)d "
+                  "bytes, got %(actual)d bytes") %
+                {'expected': image_size, 'actual': total_bytes})
 
-        success = True
-        for part in plist:
-            if not part.success:
-                success = False
+        # Check success status
+        success = all(p.success for p in plist)
+        total_size = sum(p.size for p in plist)
 
         if success:
             # Complete
+            pedict = {p.partnum: p.etag[p.partnum] for p in plist}
             mpu_list = self._get_mpu_list(pedict)
             s3_client.complete_multipart_upload(Bucket=bucket,
                                                 Key=key,
